import logging
import os
import subprocess
import threading
from lobbyboy.provider import BaseProvider
from lobbyboy.exceptions import ProviderException


logger = logging.getLogger(__name__)


class VagrantProviderException(ProviderException):
    pass


class VagrantProvider(BaseProvider):
    """
    This provider will provision new VM via local vagrant daemon.
    It won't create a real VPS, this is just a provider implementation ref.
    """

    def get_unused_vm_name(self):
        for index in range(1, 10000):
            vm_name = "lobbyboy-{}".format(index)
            if not (self.data_path / vm_name).exists():
                return vm_name
        logger.error(
            "vagrant can not found a available vm name, all id from 1 to 9999 are taken, please clean out {}".format(
                self.data_path
            )
        )
        raise Exception(
            "Can not create more VMs! Please clean out {}".format(self.data_path)
        )

    def new_server(self, chan):
        vm_name = self.get_unused_vm_name()
        vm_path = self.data_path / vm_name
        os.mkdir(str(vm_path))
        with open(str(vm_path / "Vagrantfile"), "w+") as vfile:
            file_content = self.provider_config["vagrantfile"].format(boxname=vm_name)
            vfile.write(file_content)

        over_event = threading.Event()
        self.send_timepass(chan, over_event)
        self._run_vagrant(["vagrant", "up"], cwd=str(vm_path))
        ssh_config_file = open(vm_path / "ssh_config", "wb+")
        self._run_vagrant(
            ["vagrant", "ssh-config", vm_name], cwd=str(vm_path), stdout=ssh_config_file
        )
        over_event.set()
        chan.send("New server {} created!\r\n".format(vm_name).encode())
        return vm_name, "127.0.0.1"

    def _run_vagrant(self, command_exec: list, cwd=None, stdout=None):
        logger.info("start to run command: {}".format(" ".join(command_exec)))
        capture_output = stdout is None
        vagrant_process = subprocess.run(
            command_exec, cwd=cwd, capture_output=capture_output, stdout=stdout
        )
        if vagrant_process.returncode == 0:
            logger.info(
                "vagrant_command success, command={} stdout={}, stderr={}".format(
                    " ".join(command_exec),
                    vagrant_process.stdout,
                    vagrant_process.stderr,
                )
            )
            return (
                vagrant_process.returncode,
                vagrant_process.stdout,
                vagrant_process.stderr,
            )
        logger.error(
            "vagrant_command FAILED! command={} returncode: {}, stdout: {}, stdout: {}".format(
                " ".join(command_exec),
                vagrant_process.returncode,
                vagrant_process.stdout,
                vagrant_process.stderr,
            )
        )
        raise Exception

    def ssh_server_command(self, server_id, server_ip):
        vm_path = self.data_path / server_id / "ssh_config"

        return ["ssh", "-F", str(vm_path), server_id]

    def destroy_server(self, server_id, server_ip, channel):
        vid = self._get_vagrant_machine_id(server_id)
        returncode, _, _ = self._run_vagrant(["vagrant", "destroy", "-f", vid])
        if returncode != 0:
            raise VagrantProviderException("Error when destroy {}".format(vid))

    def _get_vagrant_machine_id(self, server_id):
        _, stdout, _ = self._run_vagrant(["vagrant", "global-status"])
        stdout = stdout.decode()
        for line in stdout.split("\n"):
            if server_id in line:
                v_server_id = line.split(" ")[0]
                logger.debug(
                    "Find server_id={} by server name {}".format(v_server_id, server_id)
                )
                return v_server_id
        raise VagrantProviderException("{} not found in Vagrant!".format(server_id))



from pathlib import Path
import time
import threading


class BaseProvider:
    def __init__(self, provider_name, config, provider_config, data_path):
        self.config = config
        self.provider_name = provider_name
        self.provider_config = provider_config
        self.data_path = Path(data_path)

    def new_server(self, channel):
        """
        Args:
            channel: paramiko channel

        Returns:
            created_server_id: unique id from provision
            created_server_host: server's ip or domain address
        """
        pass

    def destroy_server(self, server_id, server_ip, channel):
        """
        Args:
            channel: Note that the channel can be None.
                     If called from server_killer, channel will be None.
                     if called when user logout from server, channel is active.
        """
        pass

    def ssh_server_command(self, server_id, server_ip):
        """
        Args:
           server_id: the server ssh to, which is returned by you from ``new_server``
           server_ip: ip or domain name.
        Returns:
            list: a command in list format, for later to run exec.
        """
        pass

    def get_bill(self):
        pass

    def send_timepass(self, chan, stop_event):
        start = time.time()

        def _print_time_elaspe():
            while not stop_event.is_set():
                current = time.time()
                chan.send(
                    "\rCreating new server... ({:.1f}s)".format(
                        current - start
                    ).encode()
                )
            chan.send(b"\r\n")

        t = threading.Thread(target=_print_time_elaspe)
        t.start()



import re
import os
import json
import logging


logger = logging.getLogger(__name__)


def parse_time_config(timestr: str):
    """
    Args:
        timestr: str, 10s, 10m, 10h, 10d
    Returns:
        seconds
    """
    matched = re.match(r"(\d+)s", timestr)
    if matched:
        return int(matched.group(1))

    matched = re.match(r"(\d+)m", timestr)
    if matched:
        return int(matched.group(1)) * 60

    matched = re.match(r"(\d+)h", timestr)
    if matched:
        return int(matched.group(1)) * 60 * 60

    matched = re.match(r"(\d+)d", timestr)
    if matched:
        return int(matched.group(1)) * 60 * 60 * 24

    raise Exception("Can not parse {}".format(timestr))


def load_server_db(db_path: str):
    """
    load from available_servers.json file, return result
    """
    try:
        server_json = json.load(open(db_path, "r+"))
    except Exception as e:
        logger.error("Error when reading available_servers.json, {}".format(str(e)))
        return []
    logger.debug(
        "open server_json, find {} available_servers: {}".format(
            len(server_json), server_json
        )
    )
    return server_json


def print_exmaple_config():
    with open(os.path.dirname(__file__) + "/conf/lobbyboy_config.toml", "r") as econf:
        print(econf.read())



import os
import json
import logging
import datetime
import paramiko
import digitalocean
import threading
import time


from lobbyboy.provider import BaseProvider
from lobbyboy.exceptions import ProviderException


DEFAULT_RSA_BITS = 3072
logger = logging.getLogger(__name__)


class NoAvailableNameException(ProviderException):
    pass


class DigitalOceanProvider(BaseProvider):
    def __init__(self, provider_name, config, provider_config, data_path):
        super().__init__(provider_name, config, provider_config, data_path)
        token = os.getenv("DIGITALOCEAN_TOKEN")
        if token:
            self.token = token
        else:
            self.token = provider_config["private_key"]

    def new_server(self, chan):
        server_name = self._generate_name()
        server_data = self.data_path / server_name
        key = self._generate_private_key(server_data)
        pubkey = "{} {}".format(key.get_name(), key.get_base64())
        ssh_keys = self.provider_config["extra_ssh_keys"]
        ssh_keys.append(pubkey)
        event = threading.Event()
        self.send_timepass(chan, event)
        droplet = digitalocean.Droplet(
            token=os.getenv("DIGITALOCEAN_TOKEN"),
            name=server_name,
            region="sgp1",  # New York 2
            image="ubuntu-20-04-x64",  # Ubuntu 20.04 x64
            size_slug="s-1vcpu-1gb",  # 1GB RAM, 1 vCPU
            ssh_keys=ssh_keys,
        )
        droplet.create()
        logger.info("create server data dir: {}".format(server_data))
        actions = droplet.get_actions()
        logger.info("create server actions: {}".format(actions))
        action1 = actions[0]
        check = 0
        while 1:
            if check <= 2:
                time.sleep(5)
            else:
                time.sleep(1)

            logger.debug("{} time check for action: {}".format(check, action1))
            check += 1
            action1.load()
            logger.debug("check result: {}".format(action1))
            # Once it shows "completed", droplet is up and running
            if action1.status == "completed":
                break

        droplet.load()
        ssh_command = self.ssh_server_command(server_name, droplet.ip_address)
        logger.info(ssh_command)
        # wait for server to start up...
        time.sleep(15)
        event.set()
        server_info_path = str(server_data / "server.json")
        self._dump_info(droplet, server_info_path)
        chan.send(
            "New server {}({}) created!\r\n".format(
                server_name, droplet.ip_address
            ).encode()
        )
        return server_name, droplet.ip_address

    def _dump_info(self, droplet, path):
        data = {
            "id": droplet.id,
            "name": droplet.name,
            "memory": droplet.memory,
            "vcpus": droplet.vcpus,
            "disk": droplet.disk,
            "region": droplet.region,
            "image": droplet.image,
            "size_slug": droplet.size_slug,
            "locked": droplet.locked,
            "created_at": droplet.created_at,
            "status": droplet.status,
            "networks": droplet.networks,
            "kernel": droplet.kernel,
            "backup_ids": droplet.backup_ids,
            "snapshot_ids": droplet.snapshot_ids,
            "action_ids": droplet.action_ids,
            "features": droplet.features,
            "ip_address": droplet.ip_address,
            "private_ip_address": droplet.private_ip_address,
            "ip_v6_address": droplet.ip_v6_address,
            "ssh_keys": droplet.ssh_keys,
            "backups": droplet.backups,
            "ipv6": droplet.ipv6,
            "private_networking": droplet.private_networking,
            "user_data": droplet.user_data,
            "volumes": droplet.volumes,
            "tags": droplet.tags,
            "monitoring": droplet.monitoring,
            "vpc_uuid": droplet.vpc_uuid,
        }
        with open(path, "w+") as serverf:
            json.dump(data, serverf)
            logger.debug("server data write to {}".format(serverf))

    def _generate_private_key(self, path):
        path = path / "id_rsa"
        rsa_key = paramiko.RSAKey.generate(DEFAULT_RSA_BITS)
        rsa_key.write_private_key_file(str(path))
        return rsa_key

    def _generate_name(self):
        datestr = datetime.datetime.now().strftime("%Y-%m-%d-%H%M")
        server_name = "lobbyboy-{}".format(datestr)
        vm_data = self.data_path / server_name
        if not vm_data.exists():
            os.mkdir(str(vm_data))
            return server_name
        raise NoAvailableNameException("Server {} already exist!".format(server_name))

    def destroy_server(self, server_id, server_ip, channel):
        logger.info("try to destroy {} {}...".format(server_id, server_ip))
        with open(self.data_path / server_id / "server.json", "r") as sfile:
            data = json.load(sfile)
        do_id = data['id']
        droplet = digitalocean.Droplet.get_object(
            api_token=self.token,
            droplet_id=do_id,
        )
        logger.info("get object from digitalocean: {}".format(droplet))
        result = droplet.destroy()
        logger.info("destroy droplet, result: {}".format(result))


    def ssh_server_command(self, server_id, server_ip):
        keypath = str(self.data_path / server_id / "id_rsa")
        command = ["ssh", "-i", keypath, "-o",  "StrictHostKeyChecking=no", "root@{}".format(server_ip)]
        logger.info("returning ssh command: {}".format(command))
        return command



import logging
import toml

logger = logging.getLogger(__name__)

def load_config(config_path):
    config = toml.load(config_path)
    logger.info("loading configs... config: {}".format(config))
    return config



#!/usr/bin/env python
from lobbyboy.server_killer import destroy_server, server_need_destroy
import time
import json
import base64
import os
import socket
import sys
import threading
import traceback
import select
import termios
import pty
import struct
import fcntl
import signal
import termios
import struct
import fcntl
import logging
import argparse
import importlib
from binascii import hexlify
from pathlib import Path

import paramiko
from paramiko import AUTH_SUCCESSFUL, AUTH_FAILED
from subprocess import Popen

from .config import load_config
from . import exceptions
from .utils import load_server_db
from .server_killer import killer
from . import __version__


DoGSSAPIKeyExchange = True
# openssh ssh-keygen: The default length is 3072 bits (RSA) or 256 bits (ECDSA).
DEFAULT_HOST_RSA_BITS = 3072
# {server_id: [Transport]}
active_session = {}
available_server_db_lock = threading.Lock()
active_sesion_lock = threading.Lock()


def setup_logs(level=logging.DEBUG):
    """send paramiko logs to a logfile,
    if they're not already going somewhere"""

    frm = "%(levelname)-.3s [%(asctime)s.%(msecs)03d] thr=%(thread)d"
    frm += " %(name)s: %(message)s"
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(frm, "%Y%m%d-%H:%M:%S"))
    logging.basicConfig(level=level, handlers=[handler])


logger = logging.getLogger(__name__)


def set_winsize(fd, row, col, xpix=0, ypix=0):
    winsize = struct.pack("HHHH", row, col, xpix, ypix)
    fcntl.ioctl(fd, termios.TIOCSWINSZ, winsize)


# TODO generate all keys when start, if key not exist.
# TODO fix server threading problems (no sleep!)


class Server(paramiko.ServerInterface):
    def __init__(self, config):
        self.pty_event = threading.Event()
        self.shell_event = threading.Event()
        self.config = config
        self.window_width = self.window_height = 0
        self.proxy_subprocess_pid = None
        self.client_exec = None
        self.client_exec_provider = None
        self.master_id = self.slave_fd = None

    def check_channel_request(self, kind, chanid):
        if kind == "session":
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_auth_password(self, username, password):
        # TODO load config file every time.
        if (username == "robey") and (password == "foo"):
            return AUTH_SUCCESSFUL
        return AUTH_FAILED

    def _get_key_class(self, key_type):
        if key_type == "ssh-rsa":
            return paramiko.RSAKey
        # TODO other types
        raise Exception("Unknown key type")

    def check_auth_publickey(self, username, key):
        key_type = key.get_name()
        logger.info("try to auth {} with key type {}...".format(username, key_type))
        reloaded_config = load_config(self.config["__config_path__"])
        keys_string = reloaded_config["user"][username]["authorized_keys"]
        key_cls = self._get_key_class(key_type)
        for line in keys_string.split("\n"):
            if not line.startswith(key_type):
                continue
            _, _key_pub = line.split(" ", 2)
            accept_key = key_cls(data=base64.b64decode(_key_pub))
            logger.info(
                "try to auth {} with key {}".format(
                    username, hexlify(accept_key.get_fingerprint()).decode()
                )
            )
            if key == accept_key:
                logger.info(
                    "accept auth {} with key {}".format(
                        username, hexlify(accept_key.get_fingerprint()).decode()
                    )
                )
                return AUTH_SUCCESSFUL
        logger.info("Can not auth {} with any key.".format(username))
        return AUTH_FAILED

    def check_auth_gssapi_with_mic(
        self, username, gss_authenticated=AUTH_FAILED, cc_file=None
    ):
        """
        .. note::
            We are just checking in `AuthHandler` that the given user is a
            valid krb5 principal! We don't check if the krb5 principal is
            allowed to log in on the server, because there is no way to do that
            in python. So if you develop your own SSH server with paramiko for
            a certain platform like Linux, you should call ``krb5_kuserok()`` in
            your local kerberos library to make sure that the krb5_principal
            has an account on the server and is allowed to log in as a user.

        .. seealso::
            `krb5_kuserok() man page
            <http://www.unix.com/man-page/all/3/krb5_kuserok/>`_
        """
        if gss_authenticated == paramiko.AUTH_SUCCESSFUL:
            return AUTH_SUCCESSFUL
        return AUTH_FAILED

    def check_auth_gssapi_keyex(
        self, username, gss_authenticated=AUTH_FAILED, cc_file=None
    ):
        # TODO
        if gss_authenticated == AUTH_SUCCESSFUL:
            logger.info("gss auth success")
            return AUTH_SUCCESSFUL
        return AUTH_FAILED

    def enable_auth_gssapi(self):
        return True

    def get_allowed_auths(self, username):
        return "gssapi-keyex,gssapi-with-mic,password,publickey"

    def check_channel_shell_request(self, channel):
        logger.info("client request shell...")
        self.pty_event.wait(10)
        if not self.pty_event.is_set():
            logger.error("Client never ask a tty, can not allocate shell...")
            raise Exception("No TTY")
        self.shell_event.set()
        return True

    def check_channel_pty_request(
        self, channel, term, width, height, pixelwidth, pixelheight, modes
    ):
        logger.info(
            "Client request pty..., term={} width={}, height={}, pixelwidth={}, pixelheight={}".format(
                term, width, height, pixelwidth, pixelwidth
            )
        )
        self.master_fd, self.slave_fd = pty.openpty()
        logger.debug(
            "user's pty ready, master_fd={}, slave_fd={}".format(
                self.master_fd, self.slave_fd
            )
        )
        self.window_width = width
        self.window_height = height
        set_winsize(self.master_fd, height, width, pixelwidth, pixelheight)
        self.pty_event.set()
        return True

    def check_channel_window_change_request(
        self, channel, width, height, pixelwidth, pixelheight
    ):
        logger.debug(
            "client send window size change reuqest... width={}, height={}, pixelwidth={}, pixelheight={}, my proxy_subprocess_pid={}, master_fd={}".format(
                width,
                height,
                pixelwidth,
                pixelheight,
                self.proxy_subprocess_pid,
                self.master_fd,
            )
        )
        self.window_width = width
        self.window_height = height

        set_winsize(self.master_fd, height, width, pixelwidth, pixelheight)
        if self.proxy_subprocess_pid is not None:
            logger.debug("send signal to {}".format(self.proxy_subprocess_pid))
            os.kill(self.proxy_subprocess_pid, signal.SIGWINCH)
        return True


def evict_active_session(transport, serverid):
    with active_sesion_lock:
        session_list = active_session.get(serverid)
        if session_list:
            active_session[serverid] = [
                _t for _t in session_list if _t.getpeername() != transport.getpeername()
            ]
        return


class SocketHandlerThread(threading.Thread):
    def __init__(self, socket_client, client_addr, config, providers) -> None:
        """
        Args:
            socket_client, client_addr: created by socket.accept()
        """
        self.socket_client = socket_client
        self.client_addr = client_addr
        self.config = config
        self.providers = providers
        self.available_server_db_path = str(
            Path(config["data_dir"]) / "available_servers.json"
        )
        super().__init__()

    def run(self):
        logger.info(
            "start new thread handle {}, addr: {}, my thread id={}".format(
                self.socket_client, self.client_addr, threading.get_ident()
            )
        )
        t = paramiko.Transport(self.socket_client, gss_kex=DoGSSAPIKeyExchange)
        try:
            t.set_gss_host(socket.getfqdn(""))
            try:
                t.load_server_moduli()
            except:
                logger.error("(Failed to load moduli -- gex will be unsupported.)")
                raise
            host_key = paramiko.RSAKey(
                filename=str(Path(self.config["data_dir"]) / "ssh_host_rsa_key")
            )
            logger.info(
                "Read host key: " + hexlify(host_key.get_fingerprint()).decode()
            )
            t.add_server_key(host_key)
            server = Server(self.config)
            try:
                t.start_server(server=server)
            except paramiko.SSHException:
                logger.error("*** SSH negotiation failed.")
                logger.error("close the transport now... {}".format(t))
                t.close()
                return

            chan = t.accept(20)
            if chan is None:
                logger.error("Client never open a new channel, close transport now...")
                t.close()
                return

            chan.send("Welcome to Lobbyboy {}!\r\n".format(__version__).encode())
            server.shell_event.wait()
            if not server.shell_event.is_set():
                logger.warn(
                    "Client never asked for a shell, I am going to end this ssh session now..."
                )
                chan.send(
                    b"*** Client never asked for a shell. Server will end session...\r\n"
                )
                t.close()
                return

            master_fd = server.master_fd
            slave_fd = server.slave_fd
            logger.info("transport peer name: {}".format(t.getpeername()))
            try:
                p, serverid = self._create_proxy_process(chan, slave_fd)
            except exceptions.UserCancelException:
                logger.warn("user input Ctrl-C or Ctrl-D during the input.")
                chan.send("Got EOF, closing session...\r\n".encode())
                chan.close()
                t.close()
                return
            except exceptions.ProviderException as e:
                logger.warn("got exceptions from provider: {}".format(str(e)))
                chan.send(
                    "Lobbyboy got exceptions from provider: {}".format(str(e)).encode()
                )
                chan.close()
                t.close()
                return

            logger.info("proxy subprocess created, pid={}".format(p.pid))
            server.proxy_subprocess_pid = p.pid

            chan.send((int(server.window_width) * "=" + "\r\n").encode())
            chan_fd = chan.fileno()
            while p.poll() is None:
                r, w, e = select.select([master_fd, chan_fd], [], [], 0.1)
                if master_fd in r:
                    d = os.read(master_fd, 10240)
                    chan.send(d)
                elif chan_fd in r:
                    o = chan.recv(10240)
                    os.write(master_fd, o)

            chan.send(
                "Lobbyboy: SSH to remote server {} closed.\r\n".format(
                    serverid
                ).encode()
            )
            evict_active_session(t, serverid)
            self.destroy_server_if_needed(serverid, chan)
            chan.shutdown(0)
            t.close()

        except Exception as e:
            logger.error("*** Caught exception: " + str(e.__class__) + ": " + str(e))
            traceback.print_exc()
            try:
                evict_active_session(t, serverid)
                t.close()
            except:
                pass

    def load_server_info(self, serverid):
        servers = load_server_db(self.available_server_db_path)
        for server in servers:
            if server["server_id"] == serverid:
                return server
        raise Exception("serverid={} not found!".format(serverid))

    def _human_time(self, seconds: int):
        return "{}s".format(seconds)

    def destroy_server_if_needed(self, serverid, channel):
        server = self.load_server_info(serverid)
        need_destroy, reason = server_need_destroy(
            active_sessions=active_session[serverid],
            serverinfo=server,
            config=self.config,
        )
        channel.send("Lobbyboy: This server {}.\r\n".format(reason).encode())
        if need_destroy:
            channel.send(
                "Lobbyboy: I will destroy {}({}) now!\r\n".format(
                    serverid, server["server_host"]
                ).encode()
            )
            provider_name = server["provider"]
            provider = self.providers[provider_name]
            destroy_server(
                provider,
                server,
                self.available_server_db_path,
                available_server_db_lock,
                channel,
            )
            channel.send("Lobbyboy: Server has been destroyed.\r\n".encode())

    def _create_proxy_process(self, channel, slave_fd):
        # if has available servers, prompt login or create
        # if no,  create, and redirect
        available_servers = load_server_db(self.available_server_db_path)
        if available_servers:
            channel.send(
                "There are {} available servers:\r\n".format(
                    len(available_servers)
                ).encode()
            )
            channel.send("{:>3} - Create a new server...\r\n".format(0))
            for index, server in enumerate(available_servers):
                channel.send(
                    "{:>3} - Enter {} {} {} ({} active sessions)\r\n".format(
                        index + 1,
                        server["provider"],
                        server["server_id"],
                        server["server_host"],
                        len(active_session.get(server["server_id"], [])),
                    )
                )
            channel.send("Please input your choice (number): ".encode())
            user_input = int(self.read_user_input_line(channel))
            if user_input == 0:
                logger.info("userinput=0, wants to create a new server...")
                serverid, serverhost, provider = self._create_new_server(channel)
            else:
                user_input -= 1
                logger.info(
                    "userinput={}, wants to ssh to an exist server...".format(
                        user_input
                    )
                )
                choosed_server = available_servers[user_input]
                serverid = choosed_server["server_id"]
                serverhost = choosed_server["server_host"]
                provider_name = choosed_server["provider"]
                provider = self.providers[provider_name]
        else:
            channel.send("There is no available servers, provision a new server...\r\n")
            serverid, serverhost, provider = self._create_new_server(channel)
        ssh_command = provider.ssh_server_command(serverid, serverhost)
        logger.info(
            "ssh to server {} {}: {}".format(
                serverid, serverhost, " ".join(ssh_command)
            )
        )
        channel.send(
            "Redirect you to {} ({})...\r\n".format(serverid, serverhost).encode()
        )
        proxy_subprocess = Popen(
            ssh_command,
            preexec_fn=os.setsid,
            stdin=slave_fd,
            stdout=slave_fd,
            stderr=slave_fd,
            universal_newlines=True,
        )
        with active_sesion_lock:
            active_session.setdefault(serverid, []).append(channel.get_transport())
        return proxy_subprocess, serverid

    def choose_providers(self, channel):
        if len(self.providers) < 1:
            raise Exception(
                "Do not have available providers to provision a new server!"
            )
        elif len(self.providers) == 1:
            return self.providers.values()[0]
        else:
            pnames = list(self.providers.keys())
            channel.send("Available VPS providers:\r\n")
            for index, name in enumerate(pnames):
                channel.send("{:>3} - {}\r\n".format(index, name))
            channel.send("Please choose a provider to create a new server: ")
            user_input = self.read_user_input_line(channel)
            choosed_provider_name = pnames[int(user_input)]
            return self.providers[choosed_provider_name]

    def _create_new_server(self, chan):
        provider = self.choose_providers(chan)
        server_id, server_ip = provider.new_server(chan)
        with available_server_db_lock:
            server_json = load_server_db(self.available_server_db_path)
            if not server_json:
                server_json = []

            server_json.append(
                {
                    "server_id": server_id,
                    "server_host": server_ip,
                    "provider": provider.provider_name,
                    "created_timestamp": time.time(),
                }
            )
            json.dump(
                server_json,
                open(Path(self.config["data_dir"]) / "available_servers.json", "w+"),
            )

        return server_id, server_ip, provider

    def read_user_input_line(self, chan):
        # TODO do not support del
        logger.debug("reading from channel... {}".format(chan))
        chars = []
        while 1:
            content = chan.recv(1)
            logger.debug("channel recv: {}".format(content))
            if content == b"\r":
                chan.send(b"\r\n")
                break
            if content == b"\x04" or content == b"\x03":
                raise exceptions.UserCancelException()
            chan.send(content)
            chars.append(content)
        return b"".join(chars).decode()


def runserver(config, providers):
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((config["listen_ip"], config["listen_port"]))
    except Exception as e:
        logger.error("*** Bind failed: " + str(e))
        traceback.print_exc()
        sys.exit(1)

    try:
        sock.listen(100)
    except Exception as e:
        logger.error("*** Listen failed: " + str(e))
        traceback.print_exc()
        sys.exit(1)
    while 1:
        logger.info("Listening for connection ...")
        try:
            client, addr = sock.accept()
        except Exception as e:
            logger.error("*** Accept new socket failed: " + str(e))
            continue
        logger.info("get a connection, from addr: {}".format(addr))
        SocketHandlerThread(client, addr, config, providers).start()


def generate_host_keys(data_dir):
    """
    check host_key exist in data_dir, if not, generate
    """
    # TODO support new key type: dss, Ed25519, ECDSA
    path = Path(data_dir) / "ssh_host_rsa_key"
    if not path.exists():
        logger.info("Host key do not exist, generate to {}...".format(str(path)))
        rsa_key = paramiko.RSAKey.generate(DEFAULT_HOST_RSA_BITS)
        rsa_key.write_private_key_file(str(path))


def load_providers(config):
    _providers = {}
    for name, pconfig in config["provider"].items():
        path = pconfig["loadmodule"]
        module_path, classname = path.split("::")
        logger.debug("loading path: {}, classname: {}".format(module_path, classname))
        module = importlib.import_module(module_path)
        provider_work_path = Path(config["data_dir"]) / name
        if not provider_work_path.exists():
            logger.info(
                "{}'s workpath {} don't exist, creating...".format(
                    name, str(provider_work_path)
                )
            )
            os.mkdir(str(provider_work_path))
        provider_obj = getattr(module, classname)(
            provider_name=name,
            config=config,
            provider_config=pconfig,
            data_path=str(provider_work_path),
        )
        _providers[name] = provider_obj

    logger.info("{} providers loaded: {}".format(len(_providers), _providers.keys()))

    return _providers


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c", "--config", dest="config_path", help="config file path", required=True
    )
    args = parser.parse_args()
    config = load_config(args.config_path)
    config["__config_path__"] = args.config_path
    log_level = logging.getLevelName(config["log_level"])
    setup_logs(log_level)
    generate_host_keys(config["data_dir"])
    providers = load_providers(config)
    killer_thread = threading.Thread(
        target=killer,
        args=(config, active_session, available_server_db_lock, providers),
        daemon=True
    )
    killer_thread.start()
    logger.info("started server_killer thread: {}".format(killer_thread))

    runserver(config, providers)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lobbyboy', 'lobbyboy.contrib', 'lobbyboy.contrib.provider']

package_data = \
{'': ['*'], 'lobbyboy': ['conf/*']}

install_requires = \
['paramiko[gssapi]>=2.8.0,<3.0.0',
 'python-digitalocean>=1.17.0,<2.0.0',
 'toml>=0.10.2,<0.11.0']

entry_points = \
{'console_scripts': ['lobbyboy-config-example = '
                     'lobbyboy.utils:print_exmaple_config',
                     'lobbyboy-server = lobbyboy.server:main']}

setup_kwargs = {
    'name': 'lobbyboy',
    'version': '0.2.1',
    'description': 'Give me a server.',
    'long_description': None,
    'author': 'laixintao',
    'author_email': 'laixintaoo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import time
import json
import pathlib
import logging


from .utils import parse_time_config, load_server_db

FIVE_MINUTES = 5 * 60
logger = logging.getLogger(__name__)


def killer(config, active_sessions, available_server_db_lock, providers):
    interval_seconds = parse_time_config(config["destroy_interval"])
    db_path = pathlib.Path(config["data_dir"]) / "available_servers.json"
    while 1:
        logger.info("killer start a new round...")
        check_all_live_servers(
            str(db_path), active_sessions, config, available_server_db_lock, providers
        )
        time.sleep(interval_seconds)


def check_all_live_servers(
    db_path: str,
    active_sessions: dict,
    config: dict,
    available_server_db_lock,
    providers,
):
    servers = load_server_db(db_path)
    for server in servers:
        serverid = server["server_id"]
        need_to_be_destroy, reason = server_need_destroy(
            active_sessions.get(serverid, []), server, config
        )
        logger.info(
            "{} need to be destroyed? {}, reson: {}".format(
                serverid, need_to_be_destroy, reason
            )
        )
        if need_to_be_destroy:
            provider_name = server["provider"]
            provider = providers[provider_name]
            destroy_server(provider, server, db_path, available_server_db_lock, None)


def server_need_destroy(active_sessions: list, serverinfo: dict, config: dict):
    """
    check if a server need to be destroyed.

    Args:
        active_session: list, track current active session on sever
        serverinfo: dict, example: {'server_id': 'lobbyboy-29', 'server_host': '127.0.0.1', 'provider': 'vagrant', 'created_timestamp': 1635175057.377654}
    Returns:
        tuple, (need_to_be_destroy: bool, reason: str)
    """
    active_session_count = len(active_sessions)
    if active_session_count:
        return False, "still have {} active sessions".format(active_session_count)
    provider_name = serverinfo["provider"]
    born_time = serverinfo["created_timestamp"]
    provider_configs = config["provider"]
    pconfig = provider_configs[provider_name]

    lived_time = time.time() - born_time
    min_life_to_live_seconds = parse_time_config(pconfig["min_life_to_live"])
    if min_life_to_live_seconds == 0:
        return True, "min_life_to_live set to 0"

    bill_time_unit = parse_time_config(pconfig["bill_time_unit"])
    destroy_spare_seconds = max(
        [parse_time_config(config["destroy_interval"]), FIVE_MINUTES]
    )
    bill_unit_left_time = (
        bill_time_unit - destroy_spare_seconds - (lived_time % bill_time_unit)
    )

    min_live_left_seconds = min_life_to_live_seconds - lived_time
    if min_live_left_seconds > 0:
        return False, "still have {} to live(min_life_to_live={}).".format(
            humanize_seconds(min_live_left_seconds), pconfig["min_life_to_live"]
        )
    if bill_unit_left_time > 0:
        return False, "still have {} to live(bill_time_unit={}).".format(
            humanize_seconds(bill_unit_left_time),
            pconfig["bill_time_unit"],
        )
    return True, "is about to enter the next billing cycle"


def humanize_seconds(seconds: int):
    if seconds <= 60:
        return "{} seconds".format(str(seconds))

    minutes = seconds // 60
    if minutes <= 60:
        return "{} minutes".format(str(minutes))

    hours = minutes // 60
    if hours <= 24:
        return "{} hours".format(str(hours))

    days = hours // 24
    return "{} days".format(str(days))


def destroy_server(
    provider,
    serverinfo,
    available_server_db_path,
    available_server_db_lock,
    channel=None,
):
    """
    Args:
        provider: provider instance, object
        serverinfo: server data dict
    """
    provider.destroy_server(serverinfo["server_id"], serverinfo["server_host"], channel)
    delete_from_server_db(
        available_server_db_lock,
        available_server_db_path,
        serverinfo["server_id"],
    )


def delete_from_server_db(available_server_db_lock, available_server_db_path, serverid):
    with available_server_db_lock:
        servers = load_server_db(available_server_db_path)
        new_servers = [s for s in servers if s["server_id"] != serverid]
        json.dump(
            new_servers,
            open(available_server_db_path, "w+"),
        )
